# import os
# import subprocess
# import tempfile
# from pathlib import Path
# import yaml
#
# from cadetrdm.container import ContainerAdapter
# from cadetrdm.batch_running import Study, Case, Options
# from cadetrdm import Environment
#
#

"""

This is a placeholder for the upcoming Apptainer Adapters

"""
# class ApptainerAdapter(ContainerAdapter):
#
#     def __init__(self):
#         pass
#
#     def run(self, yml_path):
#         with open(yml_path, "r") as stream:
#             instructions = yaml.safe_load(stream)
#
#         instructions = {key.lower(): value for key, value in instructions.items()}
#
#         study = Study(**instructions["study"], suppress_lfs_warning=True)
#         options = Options(**instructions["options"])
#         environment = Environment(**instructions["environment"])
#         case = Case(study, options, environment)
#
#         return self.run_case(case, command=instructions["command"])
#
#     def run_case(self, case: Case, command: str = None):
#         if self.image is None:
#             image = self._build_image(case)
#         else:
#             image = self.image
#
#         container_tmp_filename = "/tmp/options.json"
#         options_tmp_filename = self._dump_options(case)
#
#         full_command = self._prepare_case_command(
#             case=case,
#             command=command,
#             container_tmp_filename=container_tmp_filename
#         )
#
#         log, return_code = self._run_command(
#             container_tmp_filename=container_tmp_filename,
#             full_command=full_command,
#             image=image,
#             options_tmp_filename=options_tmp_filename
#         )
#
#         return log, return_code
#
#     def _run_command(self, container_tmp_filename, full_command, image, options_tmp_filename):
#
#         ssh_location = Path.home() / ".ssh"
#         if not ssh_location.exists():
#             raise FileNotFoundError("No ssh folder found. Please report this on GitHub/CADET/CADET-RDM")
#
#         container = self.client.containers.run_yml(
#             image=image,
#             command=full_command,
#             volumes={
#                 f"{Path.home()}/.ssh": {'bind': "/root/.ssh_host_os", 'mode': "ro"},
#                 options_tmp_filename.absolute().as_posix(): {'bind': container_tmp_filename, 'mode': 'ro'}
#             },
#             detach=True,
#             remove=False
#         )
#
#         full_log = []
#         # Step 2: Attach to the container's logs
#         for log in container.logs(stream=True):
#             full_log.append(log.decode("utf-8"))
#             print(log.decode("utf-8"), end="")
#
#         # Wait for the container to finish execution
#         result = container.wait()
#         exit_code = result["StatusCode"]
#
#         container.remove()
#
#         return full_log, exit_code
#
#     def _dump_options(self, case):
#         tmp_filename = Path("tmp/" + next(tempfile._get_candidate_names()) + ".json")
#         case.options.dump_json_file(tmp_filename)
#         return tmp_filename
#
#     def _build_image(self, case):
#         cwd = os.getcwd()
#         with open(case.project_repo.path / "Dockerfile", "rb") as dockerfile:
#             os.chdir(case.project_repo.path.as_posix())
#
#             image, logs = self.client.images.build(
#                 path=case.project_repo.path.as_posix(),
#                 # fileobj=dockerfile,  # A file object to use as the Dockerfile.
#                 tag=case.project_repo.name + ":" + case.name[:10],  # A tag to add to the final image
#                 quiet=False,  # Whether to return the status
#                 pull=True,  # Downloads any updates to the FROM image in Dockerfiles
#
#             )
#         if case.options.debug:
#             for log in logs:
#                 print(log)
#         os.chdir(cwd)
#         return image
#
#     def pull_image(self, repository, tag=None, all_tags=False, **kwargs):
#         self.image = self.client.images.pull(
#             repository=repository,
#             tag=tag,
#             all_tags=all_tags,
#             **kwargs
#         )
#
#     def _push_image(self, repository, tag=None, **kwargs):
#         self.client.images.push(
#             repository=repository,
#             tag=tag,
#             **kwargs
#         )
#
#     def _tag_image(self, image, repository, tag=None, **kwargs):
#         """
#         Tag this image into a repository. Similar to the ``docker tag``
#         command.
#
#         Args:
#             repository (str): The repository to set for the tag
#             tag (str): The tag name
#             force (bool): Force
#
#         Raises:
#             :py:class:`docker.errors.APIError`
#                 If the server returns an error.
#
#         Returns:
#             (bool): ``True`` if successful
#         """
#         image.tag(repository=repository, tag=tag, **kwargs)
#         return image
#
#     def build_and_push_image(self, case, repository, tag=None, **kwargs):
#         image = self._build_image(case)
#         image = self._tag_image(image, repository, tag, **kwargs)
#         self._push_image(repository, tag, **kwargs)
#
#     def _update_Dockerfile_with_env_reqs(self, case):
#         case.project_repo._reset_hard_to_head(force_entry=True)
#
#         dockerfile = Path(case.project_repo.path) / "Dockerfile"
#         install_command = case.environment.prepare_install_instructions()
#         if install_command is None:
#             return
#
#         with open(dockerfile, "a") as handle:
#             handle.write(f"\n{install_command}\n")
#
#     def __del__(self):
#         self.client.close()



from pathlib import Path
import pytest

from cadetrdm import Study, Options, Environment, Case, ProjectRepo
from cadetrdm.container import DockerAdapter


@pytest.mark.container
def test_run_dockered():
    WORK_DIR = Path.cwd() / "tmp"
    WORK_DIR.mkdir(parents=True, exist_ok=True)

    rdm_example = ProjectRepo(
        path=WORK_DIR / 'template',
        url="git@github.com:ronald-jaepel/rdm_testing_template.git",
        suppress_lfs_warning=True
    )

    options = Options()
    options.debug = False
    options.push = False
    options.commit_message = 'Trying out new things'
    options.optimizer_options = {
        "optimizer": "U_NSGA3",
        "pop_size": 2,
        "n_cores": 2,
        "n_max_gen": 1,
    }

    matching_environment = Environment(
        conda_packages={
            "libsqlite": "==3.48.0"
        },
        # pip_packages={
        #     "cadet-rdm": "git+https://github.com/cadet/CADET-RDM.git@3e073dd85c5e54d95422c0cdcc1190d80da9e138"
        # }
    )

    case = Case(project_repo=rdm_example, options=options, environment=matching_environment)
    docker_adapter = DockerAdapter()
    has_run_study = case.run_study(container_adapter=docker_adapter, force=True)
    assert has_run_study

    options.optimizer_options = {
        "optimizer": "NOT_AN_OPTIMIZER",
        "pop_size": 2,
        "n_cores": 2,
        "n_max_gen": 1,
    }

    case = Case(project_repo=rdm_example, options=options, environment=matching_environment)
    docker_adapter = DockerAdapter()
    has_run_study = case.run_study(container_adapter=docker_adapter, force=True)
    assert not has_run_study


@pytest.mark.container
def test_dockered_from_yml():
    docker_adapter = DockerAdapter()
    has_run_study = docker_adapter.run_yml((Path(__file__).parent.resolve() / "case.yml").as_posix())
    assert has_run_study



__version__ = "1.0.1"


from cadetrdm.conda_env_utils import prepare_conda_env
from cadetrdm.options import Options
from cadetrdm.repositories import ProjectRepo, JupyterInterfaceRepo
from cadetrdm.initialize_repo import initialize_repo
from cadetrdm.environment import Environment
from cadetrdm.batch_running import Study, Case
from cadetrdm.wrapper import tracks_results
from cadetrdm.tools.process_example import process_example



import subprocess
from pathlib import Path
import shlex

import click

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass


@cli.command(help="Create an empty CADET-RDM repository or initialize over an existing git repo.")
@click.option('--output_repo_name', default="output",
              help='Name of the folder where the tracked output should be stored. Optional. Default: "output".')
@click.option('--gitignore', default=None,
              help='List of files to be added to the gitignore file. Optional.')
@click.option('--gitattributes', default=None,
              help='List of files to be added to the gitattributes file. Optional.')
@click.option('--cookiecutter', default=None,
              help='URL or path to cookiecutter template. Optional.')
@click.argument('path_to_repo', required=False)
def init(path_to_repo: str = None, output_repo_name: (str | bool) = "output", gitignore: list = None,
         gitattributes: list = None, cookiecutter: str = None, output_repo_kwargs: dict = None):
    if path_to_repo is None:
        path_to_repo = "."
    from cadetrdm.initialize_repo import initialize_repo as initialize_git_repo_implementation
    initialize_git_repo_implementation(path_to_repo, output_repo_name, gitignore,
                                       gitattributes, output_repo_kwargs, cookiecutter)


@cli.command(help="Clone a repository into a new d^irectory.")
@click.argument('project_url')
@click.argument('dest', required=False)
def clone(project_url, dest: str = None):
    from cadetrdm import ProjectRepo
    repo = ProjectRepo.clone(url=project_url, to_path=dest)
    del repo


@cli.command(name="log", help="Show commit logs.")
def print_log():
    from cadetrdm.repositories import BaseRepo

    repo = BaseRepo(".")
    repo.print_log()
    del repo


@cli.command(name="check", help="Ensure metadata is consistent.")
def check():
    repo = get_project_repo()
    repo.check()
    del repo


@cli.command(help="Push all changes to the project and output repositories.")
@click.option('--single', "-s", is_flag=True, help="Push only changes of the current branch.")
def push(single=False):
    repo = get_project_repo()
    repo.push(push_all=not single)
    del repo


def get_project_repo(path: str = "."):
    """
    Get the project repo to a given path
    :return:
    """
    from cadetrdm.repositories import ProjectRepo, BaseRepo
    base_repo = BaseRepo(path)
    if base_repo._metadata["is_project_repo"]:
        repo = ProjectRepo(path)
    elif base_repo._metadata["is_output_repo"]:
        repo = ProjectRepo(Path(path).parent)
    else:
        raise ValueError("Current directory is neither Project nor Output repository")
    return repo


@cli.command(help="Record changes to the repository")
@click.option("--message", "-m", help="commit message")
@click.option("--all", "-a", is_flag=True, help="commit all changed files")
def commit(message, all):
    from cadetrdm.repositories import ProjectRepo
    repo = ProjectRepo(".")
    repo.commit(message, all)
    del repo


@cli.group(help="Execute commands and track the results.")
def run():
    pass


@run.command(name="python")
@click.argument('file_name')
@click.argument('results_commit_message')
def run_python_file(file_name, results_commit_message):
    from cadetrdm.repositories import ProjectRepo
    repo = ProjectRepo(".")
    repo.enter_context()
    subprocess.run(["python", file_name])
    repo.exit_context(results_commit_message, {"command": f"python {file_name}"})
    del repo


@run.command(name="command")
@click.argument('command', nargs=-1)
@click.argument('results_commit_message')
def run_command(command, results_commit_message):
    from cadetrdm.repositories import ProjectRepo
    repo = ProjectRepo(".")
    repo.enter_context(force=True)
    if isinstance(command, tuple):
        command = " && ".join(command)
    command = shlex.split(command)
    results = subprocess.run(command, text=True, shell=False, capture_output=True)
    print(f"Command completed with stdout: {results.stdout} \nand stderr: {results.stderr}")
    repo.exit_context(results_commit_message, {"command": command})
    del repo


@run.command(name="dockered")
@click.argument('yaml_path')
def run_dockered(yaml_path):
    from cadetrdm.container import DockerAdapter
    docker_adapter = DockerAdapter()
    docker_adapter.run(yaml_path)


@cli.group(help="Create, add, and manage remotes.")
def remote():
    pass


@remote.command(name="add", help="Add")
@click.option('--name', '-n', default=None)
@click.argument('remote_url')
def add_remote(name: str = None, remote_url: str = None):
    from cadetrdm.repositories import BaseRepo
    repo = BaseRepo(".")
    repo.add_remote(remote_url=remote_url, remote_name=name)
    print("Done.")
    del repo


@remote.command(name="set-url", help="Add")
@click.argument('name')
@click.argument('remote_url')
def set_url(name: str, remote_url: str):
    from cadetrdm.repositories import BaseRepo
    repo = BaseRepo(".")
    repo.remote_set_url(url=remote_url, name=name)
    print(f"Set url of remote {name} to {remote_url}. Commiting changes to metadata.")
    project_repo = get_project_repo()
    project_repo.check(commit=True)
    del repo, project_repo


@remote.command(name="create")
@click.argument('url')
@click.argument('namespace')
@click.argument('name')
@click.argument('username', required=False)
@click.argument('push', required=False)
def create_remotes(url, namespace, name, username=None, push=True):
    if username is None:
        username = namespace

    from cadetrdm.repositories import ProjectRepo
    repo = ProjectRepo(".")
    repo.create_remotes(name=name, namespace=namespace, url=url, username=username, push=push)
    del repo


@remote.command(name="list")
def list_remotes():
    from cadetrdm.repositories import BaseRepo
    repo = BaseRepo(".")
    for _remote, url in zip(repo.remotes, repo.remote_urls):
        print(_remote, ": ", url)
    del repo


@cli.group(help="Manage large file storage settings.")
def lfs():
    pass


@lfs.command(name="add", help="Add a filetype to git lfs.")
@click.argument('file_types', nargs=-1)
def add_filetype_to_lfs(file_types: list, ):
    from cadetrdm.repositories import OutputRepo
    repo = OutputRepo(".")
    for f_type in file_types:
        repo.add_filetype_to_lfs(f_type)
    del repo


@cli.group(help="Manage data and input-data-repositories.")
def data():
    pass


@data.command(name="import", help="Import static data into the output repository without commiting the project status.")
@click.argument('source_path')
@click.argument('commit_message')
def import_remote_repo(source_path, commit_message):
    from cadetrdm.repositories import ProjectRepo
    repo = ProjectRepo(".")
    repo.import_static_data(
        source_path=source_path,
        commit_message=commit_message,
    )
    del repo


@data.command(name="clone", help="Import a remote repository into a given location.")
@click.argument('source_repo_location')
@click.argument('source_repo_branch')
@click.argument('target_repo_location', required=False)
def import_remote_repo(source_repo_location, source_repo_branch, target_repo_location=None):
    from cadetrdm.repositories import BaseRepo
    repo = BaseRepo(".")
    repo.import_remote_repo(source_repo_location=source_repo_location,
                            source_repo_branch=source_repo_branch,
                            target_repo_location=target_repo_location)
    del repo


@data.command(name="fetch", help="Fill data cache based on cadet-rdm.json.")
@click.option('--re_load', is_flag=True,
              help='Re-load all data.')
def fill_data_from_cadet_rdm_json(re_load=False):
    from cadetrdm.repositories import ProjectRepo
    repo = ProjectRepo(".")
    repo.fill_data_from_cadet_rdm_json(re_load=re_load)
    del repo


@data.command(name="cache", help="Copy data from the output repo to the cache.")
@click.argument("branch")
def copy_to_cache(branch: str):
    from cadetrdm.repositories import ProjectRepo
    repo = ProjectRepo(".")
    repo.copy_data_to_cache(branch)
    del repo


@data.command(name="verify", help="Verify that cache is unchanged.")
def verify_unchanged_cache():
    from cadetrdm.repositories import BaseRepo
    repo = BaseRepo(".")
    repo.verify_unchanged_cache()
    del repo


@data.command(name="log", help="Print data logs.")
def print_data_log():
    from cadetrdm.repositories import ProjectRepo, BaseRepo, OutputRepo
    import json

    repo = BaseRepo(".")

    with open(repo.data_json_path, "r") as handle:
        rdm_data = json.load(handle)
    if rdm_data["is_project_repo"]:
        repo = ProjectRepo(".")
        repo.print_output_log()
    else:
        repo = OutputRepo()
        repo.print_output_log()
    del repo



from abc import abstractmethod
import subprocess


class ContainerAdapter:
    def __init__(self, image=None):
        self.image = image
        return

    @abstractmethod
    def run_case(self, case, command):
        return

    @staticmethod
    def _prepare_base_commands():
        # copy over git config
        git_config_list = subprocess.check_output(
            "git config --list --show-origin --global",
            shell=True
        ).decode().split("\n")
        git_config = {
            "user.name": None,
            "user.email": None,
        }
        for line in git_config_list:
            for key in git_config.keys():
                if key in line:
                    value = line.split("=")[-1]
                    # print(value)
                    git_config[key] = value

        commands_git = [f'git config --global {key} "{value}"' for key, value in git_config.items()]

        # ensure ssh in the container knows where to look for known_hosts and that .ssh/config is read-only
        command_ssh = 'cp -r /root/.ssh_host_os/. /root/.ssh && chmod 600 /root/.ssh/.'

        commands = commands_git + [command_ssh, ]
        return commands

    @staticmethod
    def _prepare_case_command(case, command, container_options_filename):
        commands = ContainerAdapter._prepare_base_commands()

        command_install = case.environment.prepare_install_instructions()
        if command_install is not None:
            commands.append(command_install)

        # pull the study from the URL into a "study" folder
        command_pull = f"rdm clone {case.project_repo.url} study"
        # cd into the "study" folder
        command_cd = "cd study"
        # checkout branch
        command_checkout = f"git checkout {case.project_repo.active_branch}"
        # run_yml main.py with the options, assuming main.py lies within a sub-folder with the same name as the study.name
        if command is None:
            command_python = f"python {case.project_repo.name}/main.py {container_options_filename}"
        else:
            command_python = command

        commands.extend([command_pull, command_cd, command_checkout, command_python])
        full_command = ' && '.join(commands)
        return full_command



import argparse
import os
import shutil
from abc import abstractmethod
from pathlib import Path
import subprocess

from cadetrdm import ProjectRepo, Options
from cadetrdm.wrapper import tracks_results


class ParallelizationBase:
    def __init__(self, n_cores=1):
        self.n_cores = n_cores

    @abstractmethod
    def run(self, func, args_list):
        """Run a function over a list of input args and return the output.

        Parameters
        ----------
        func : callable
            The function wrapping the shell command.
        args_list : list | iterable
            List of args to the func.

        Returns
        -------
        List of function returns
        """
        return


class SequentialBackend(ParallelizationBase):
    def run(self, func, args_list):
        results = []
        for args in args_list:
            results.append(func(*args))
        return results

#
class JoblibBackend(ParallelizationBase):
    def run(self, func, args_list):
        from joblib import Parallel, delayed
        Parallel(n_jobs=self.n_cores)(delayed(func)(*args) for args in args_list)


class PathosBackend(ParallelizationBase):
    def run(self, func, args_list):
        import pathos
        with pathos.pools.ProcessPool(ncpus=self.n_cores) as pool:
            # *zip(*args_list) because for multiple args, pathos expects (func, args1_list, args2_list)
            results = pool.map(func, *zip(*args_list))
        return results


def run_func_over_args_list(func, args_list, backend=None, n_cores=1):
    """Run a function over a list of input args and return the output.

    Parameters
    ----------
    func : callable
        The function wrapping the shell command.
    args_list : list | iterable
        List of tuple of args to the func.
    backend : ParallelizationBase, optional
        Backend for parallelization. Default is SequentialBackend
    n_cores : int, optional
        Number of cores to use for parallelization

    Returns
    -------
    List of function returns
    """
    if type(args_list[0]) not in (list, tuple):
        args_list = [(x,) for x in args_list]

    if backend is None and n_cores == 1:
        backend = SequentialBackend()
    else:
        backend = PathosBackend(n_cores=n_cores)

    return backend.run(func, args_list)


def run_command(command):
    """Run a shell command and return its output.

    Parameters
    ----------
    command : str
        The shell command to run_yml.

    Returns
    -------
    None
    """
    print(command)
    return subprocess.run(command, shell=True, check=True)


def remove_non_jupytext_files(file_list):
    """
    Filter out all python files that do not have a 'jupytext:' header section.

    :param file_list:
    :return:
    """
    if len(file_list) == 0:
        raise RuntimeError("No python files found in repository")
    filtered_list = []
    for file in file_list:
        with open(file, "r", encoding="utf8") as handle:
            lines = handle.readlines()
        if any("jupytext:" in line for line in lines):
            filtered_list.append(file)
    if len(file_list) == 0:
        raise RuntimeError("All python files found in repository are not tagged with jupytext.")
    return filtered_list


def create_output(root_path: Path, output_path: Path, n_cores=1):
    """Create solution files.

    Parameters
    ----------
    root_path : Path
        Root directory wherein all .py files are located
    output_path : Path
        Root directory where all compiled .ipynbs should be placed
    n_cores : int, optional
        Number of cpu cores to use for parallelization

    Returns
    -------
    None
    """
    if os.path.exists(output_path):
        shutil.rmtree(output_path)
    shutil.copytree(root_path, output_path)

    # Find all myst files recursively
    python_files = list(output_path.glob("**/*.py"))

    # Filter out checkpoints
    python_files = [
        file for file in python_files if (".ipynb_checkpoints" not in file.as_posix())
    ]

    # Filter out python files that are not marked for jupytext conversion
    python_files = remove_non_jupytext_files(python_files)

    # Make ipynbs
    ipynb_files = [file.with_suffix(".ipynb") for file in python_files]

    # Run jupytext for each myst file
    run_func_over_args_list(
        func=convert_python_to_ipynb,
        args_list=[
            (python_path.as_posix(), ipynb_path.as_posix())
            for python_path, ipynb_path in zip(python_files, ipynb_files)
        ],
        n_cores=n_cores,
    )


def convert_python_to_ipynb(myst_file_path, ipynb_file_path):
    """Run jupytext with --output ipynb flag on myst_file_path. Will skip README files.
    Will run_yml the jupyter notebooks.

    Parameters
    ----------
    myst_file_path : str
        path to myst file as posix.

    ipynb_file_path : str
        path to output ipynb file as posix.

    Returns
    -------
    None
    """
    if "README" in myst_file_path:
        return
    results = run_command(
        f'jupytext --output "{ipynb_file_path}" --execute "{myst_file_path}"'
    )
    return results


@tracks_results
def process_example(repo: ProjectRepo, options, **kwargs):
    """Run post-commit tasks based on command-line arguments.

    Returns
    -------
    None
    """
    parser = argparse.ArgumentParser(description="Perform post-commit tasks.")
    parser.add_argument("--n_cores", help="Number of cores to use.")

    args = parser.parse_args()

    # This isn't great, but for now (and with argparse) the best I could think of
    for kwarg_key, kwarg_value in kwargs.items():
        if kwarg_value is None:
            continue
        args.__setattr__(kwarg_key, kwarg_value)

    if args.n_cores is None:
        args.n_cores = 1

    create_output(
        root_path=repo.path / options.source_directory,
        output_path=repo.output_path / options.source_directory,
        n_cores=args.n_cores,
    )


"""Example usage:
    options = Options()
    options.commit_message = "Process example"
    options.debug = False
    options.push = True
    options.source_directory = "src"
    main(options)
"""



# from pathlib import Path
# import pytest
#
# from cadetrdm import Study, Options, Environment, Case
# from cadetrdm.container import ApptainerAdapter
#
#
# @pytest.mark.container
# def test_run_apptainer():
#     WORK_DIR = Path.cwd() / "tmp"
#     WORK_DIR.mkdir(parents=True, exist_ok=True)
#
#     rdm_example = Study(
#         WORK_DIR / 'template',
#         "git@github.com:ronald-jaepel/rdm_testing_template.git",
#         suppress_lfs_warning=True
#     )
#
#     options = Options()
#     options.debug = False
#     options.push = False
#     options.commit_message = 'Trying out new things'
#     options.optimizer_options = {
#         "optimizer": "U_NSGA3",
#         "pop_size": 2,
#         "n_cores": 2,
#         "n_max_gen": 1,
#     }
#
#     matching_environment = Environment(
#         pip_packages={
#             "cadet-rdm": "git+https://github.com/cadet/CADET-RDM.git@3e073dd85c5e54d95422c0cdcc1190d80da9e138"
#         }
#     )
#
#     case = Case(study=rdm_example, options=options, environment=matching_environment)
#     docker_adapter = ApptainerAdapter()
#     has_run_study = case.run_study(container_adapter=docker_adapter, force=True)
#     assert has_run_study
#
#     options.optimizer_options = {
#         "optimizer": "NOT_AN_OPTIMIZER",
#         "pop_size": 2,
#         "n_cores": 2,
#         "n_max_gen": 1,
#     }
#
#     case = Case(study=rdm_example, options=options, environment=matching_environment)
#     docker_adapter = ApptainerAdapter()
#     has_run_study = case.run_study(container_adapter=docker_adapter, force=True)
#     assert not has_run_study
#
#
# @pytest.mark.container
# def test_dockered_from_yml():
#     docker_adapter = ApptainerAdapter()
#     has_run_study = docker_adapter.run((Path(__file__).parent.resolve() / "case.yml").as_posix())
#     assert has_run_study



import traceback
import warnings
from pathlib import Path
import subprocess
from typing import Dict

# from cadetrdm.container.containerAdapter import ContainerAdapter
from cadetrdm.repositories import ProjectRepo
from cadetrdm import Options
from cadetrdm.environment import Environment
from cadetrdm.logging import LogEntry


class Case:
    def __init__(self, project_repo: ProjectRepo = None, options: Options = None, environment: Environment = None,
                 name: str = None,
                 study=None):
        if study is not None:
            warnings.warn(
                "Initializing Case() with the study= kwarg is deprecated and will be removed in the future. "
                "Please use project_repo=",
                FutureWarning
            )
            project_repo = study

        if name is None:
            name = project_repo.name + "_" + options.get_hash()

        self.name = name

        self.project_repo = project_repo
        self.options = options
        self.environment = environment
        self._options_hash = options.get_hash()
        self.results_branch = None
        self._current_environment = None

    def __str__(self):
        return self.name

    @property
    def status_file(self):
        return Path(self.project_repo.path).parent / (Path(self.project_repo.path).name + ".status")

    @property
    def status(self):
        status, _ = self._read_status()
        return status

    @status.setter
    def status(self, status):
        """Update the status file with the current execution status."""

        with open(self.status_file, "w") as f:
            f.write(f"{status}@{self.project_repo.current_commit_hash}")

    @property
    def status_hash(self):
        _, status_hash = self._read_status()
        return status_hash

    def _read_status(self):
        """Check the status of the study and decide whether to proceed.

        Args:
            repo_path (Path): The path to the repository containing the status file.

        Returns:
            tuple: A tuple containing the status string and the current hash,
            or None, None if the status cannot be determined.
        """

        if not self.status_file.exists():
            return None, None

        with open(self.status_file) as f:
            status = f.read().strip()
            try:
                status, current_hash = status.split("@")
            except ValueError as e:
                if status == '':
                    return None, None
                else:
                    raise e

            return status, current_hash

    @property
    def is_running(self, ):
        if self.status == 'running':
            return True

        return False

    @property
    def has_results_for_this_run(self):
        self.results_branch = self._get_results_branch()
        if self.results_branch is None:
            return False
        else:
            return True

    def _get_results_branch(self):
        """
        Search the output log for an entry (i.e. a results branch) with matching study commit hash and options hash.
        If environment is given, the environment is also enforced.

        :return:
        str: name of the results branch or None.
        """
        output_log: Dict[str, LogEntry] = self.project_repo.output_log.entries
        options_hash = self.options.get_hash()
        study_hash = self.project_repo.current_commit_hash

        found_results_with_incorrect_study_hash = False
        found_results_with_incorrect_options = False
        found_results_with_incorrect_environment = False

        semi_correct_hits = []
        for output_branch, log_entry in reversed(output_log.items()):
            matches_study_hash = log_entry.matches_study_hash(study_hash)
            matches_options_hash = log_entry.matches_options_hash(options_hash)
            matches_environment = log_entry.fulfils_environment(self.environment)

            if matches_study_hash and matches_options_hash and matches_environment:
                return log_entry.output_repo_branch

            elif matches_study_hash and not matches_options_hash and matches_environment:
                found_results_with_incorrect_options = True
                semi_correct_hits.append(
                    f"Found matching study commit hash {study_hash[:7]}, but incorrect options hash "
                    f"(needs: {options_hash[:7]}, has: {log_entry.options_hash[:7]})"
                )

            elif not matches_study_hash and matches_options_hash and matches_environment:
                found_results_with_incorrect_study_hash = True
                semi_correct_hits.append(
                    f"Found matching options hash  {options_hash[:7]}, but incorrect study commit hash "
                    f"(needs: {study_hash[:7]}, has: {log_entry.project_repo_commit_hash[:7]})"
                )
            elif matches_study_hash and matches_options_hash and not matches_environment:
                found_results_with_incorrect_environment = True
                semi_correct_hits.append(
                    f"Found matching options hash  {options_hash[:7]}, matching study commit hash "
                    f"{study_hash[:7]}, but wrong environment specification."
                )

        if found_results_with_incorrect_study_hash:
            [print(line) for line in semi_correct_hits]
            print(
                "No matching results were found for this study version, but results with these options were found for "
                "other study versions. Did you recently update the study?"
            )
        elif found_results_with_incorrect_options:
            [print(line) for line in semi_correct_hits]
            print(
                "No matching results were found for these options, but results with other options were found for "
                "this study versions. Did you recently change the options?"
            )
        elif found_results_with_incorrect_environment:
            [print(line) for line in semi_correct_hits]
            print(
                "No matching results were found for this environment, but results with other environments were "
                "found for this study versions."
            )

        else:
            print("No matching results were found for these options and study version.")
        return None

    def run_study(self, force=False, container_adapter: "ContainerAdapter" = None, command: str = None) -> bool:
        """
        Run specified study commands in the given repository.

        :returns
            boolean indicating if the results for this case are available, either pre-computed or computed now.

        """
        if not force and self.is_running:
            print(f"{self.project_repo.name} is currently running. Skipping...")
            return False

        print(f"Running {self.name} in {self.project_repo.path} with: {self.options}")
        if not self.options.debug:
            self.project_repo.update()
        else:
            print("WARNING: Not updating the repositories while in debug mode.")

        if self.has_results_for_this_run and not force:
            print(f"{self.project_repo.path} has already been computed with these options. Skipping...")
            return True

        if container_adapter is None and self.can_run_study is False:
            print(f"Current environment does not match required environment. Skipping...")
            self.status = 'failed'
            return False

        try:
            self.status = 'running'

            if container_adapter is not None:
                log, return_code = container_adapter.run_case(self, command=command)
                if return_code != 0:
                    self.status = "failed"
                    return False
            else:
                self.project_repo.module.main(self.options, str(self.project_repo.path))

            print("Command execution successful.")
            self.status = 'finished'
            return True

        except (KeyboardInterrupt, Exception) as e:
            traceback.print_exc()
            self.status = 'failed'
            return False

    @property
    def can_run_study(self) -> bool:
        return self.environment is None or self._check_execution_environment()

    def _check_execution_environment(self):
        if self._current_environment is None:
            existing_environment = subprocess.check_output(f"conda env export", shell=True).decode()
            self._current_environment = Environment.from_yml_string(existing_environment)

        return self._current_environment.fulfils_environment(self.environment)

    @property
    def _results_path(self):
        if self.results_branch is None:
            return None
        else:
            return self.project_repo.cache_folder_for_branch(self.results_branch)

    def load(self, ):
        if self.results_branch is None or self.options.get_hash() != self._options_hash:
            self.results_branch = self._get_results_branch()
            self._options_hash = self.options.get_hash()

        if self.results_branch is None:
            print(f"No results available for Case({self.project_repo.path, self.options.get_hash()[:7]})")
            return None

        if self._results_path.exists():
            return

        self.project_repo.copy_data_to_cache(self.results_branch)

    @property
    def results_path(self):
        self.load()

        return self._results_path



from functools import wraps
from pathlib import Path
from copy import deepcopy

from cadetrdm.repositories import ProjectRepo
from cadetrdm import Options


def tracks_results(func):
    """
    Tracks results using CADET-RDM.
    Adds the project_repo to the function arguments and adds the output_branch_name to the return information.

    """

    @wraps(func)
    def wrapper(options, repo_path='.'):
        if type(options) is str and Path(options).exists():
            options = Options.load_json_file(options)
        elif type(options) is str:
            options = Options.load_json_str(options)
        if type(options) is dict:
            options = Options(options)

        for key in ["commit_message", "debug"]:
            if key not in options or options[key] is None:
                raise ValueError(f"Key {key} not found in options. Please supply options.{key}")

        if options.get_hash() != Options.load_json_str(options.dump_json_str()).get_hash():
            raise ValueError("Options are not serializable. Please only use python natives and numpy ndarrays.")

        project_repo = ProjectRepo(repo_path, options=options)

        project_repo.options_hash = options.get_hash()

        with project_repo.track_results(
                options.commit_message,
                debug=options.debug,
                force=True
        ) as new_branch_name:
            options.dump_json_file(project_repo.output_path / "options.json")
            results = func(project_repo, options)

        if not options.debug and "push" in options and options["push"]:
            project_repo.push()

        return new_branch_name, results

    return wrapper



import os
import subprocess
import tempfile
from pathlib import Path

import yaml

from cadetrdm.container import ContainerAdapter
from cadetrdm.batch_running import Case
from cadetrdm import Environment, ProjectRepo, Options


class PodmanAdapter(ContainerAdapter):

    def run(self, command, mounts=None):
        commands = self._prepare_base_commands()
        commands.append(command)
        full_command = " && ".join(commands)

        log, return_code = self._run_command(
            full_command=full_command,
            image=self.image,
            mounts=mounts,
        )

        return log, return_code

    def run_yml(self, yml_path):
        with open(yml_path, "r") as stream:
            instructions = yaml.safe_load(stream)

        instructions = {key.lower(): value for key, value in instructions.items()}

        project_repo = ProjectRepo(**instructions["projectrepo"], suppress_lfs_warning=True)
        options = Options(**instructions["options"])
        environment = Environment(**instructions["environment"])
        case = Case(project_repo, options, environment)

        return self.run_case(case=case, command=instructions["command"])

    def run_case(self, case: Case, command: str = None):
        if self.image is None:
            raise ValueError("Please first specify an image name for the ContainerAdapter to use")

        container_tmp_filename = "/tmp/options.json"
        options_tmp_filename = self._dump_options(case)

        full_command = self._prepare_case_command(
            case=case,
            command=command,
            container_options_filename=container_tmp_filename
        )

        log, return_code = self._run_command(
            full_command=full_command,
            image=self.image,
            mounts={options_tmp_filename: container_tmp_filename},
        )
        return log, return_code

    def _run_command(self, full_command, image, mounts=None):
        """

        :param full_command:
        :param image:
        :param mounts: Dictionary mapping host paths to container paths
        :return:
        """

        ssh_location = Path.home() / ".ssh"
        if not ssh_location.exists():
            raise FileNotFoundError("No ssh folder found. Please report this on GitHub/CADET/CADET-RDM")

        full_command = full_command.replace('"', "'")

        volume_mounts = ""
        if mounts is None:
            mounts = {}
        for host_path, container_path in mounts.items():
            volume_mounts += f'-v {host_path.absolute().as_posix()}:{container_path}:ro '

        podman_command = (
            f'podman run '
            '--rm '  # remove container after run_yml (to keep space usage low)
            f'-v {ssh_location}:/root/.ssh_host_os:ro '  # mount ssh folder for the container to access
            f'{volume_mounts}'  # mount options file
            f'{image} '  # specify image name
            f'bash -c "{full_command}"'  # run_yml command in bash shell
        )

        result = subprocess.run(podman_command, shell=True, capture_output=True)

        full_log = result.stdout.decode() + result.stderr.decode()
        exit_code = result.returncode
        print(full_log)
        print(f"RETURN CODE: {exit_code}")

        return full_log, exit_code

    def _dump_options(self, case):
        if not Path("tmp").exists():
            os.makedirs("tmp")
        tmp_filename = Path("tmp/" + next(tempfile._get_candidate_names()) + ".json")
        case.options.dump_json_file(tmp_filename)
        return tmp_filename

    # def _build_image(self, case):
    #     raise NotImplementedError
    #
    # def _push_image(self, repository, tag=None, **kwargs):
    #     raise NotImplementedError
    #
    # def _tag_image(self, image, repository, tag=None, **kwargs):
    #     """
    #     Tag this image into a repository. Similar to the ``docker tag``
    #     command.
    #
    #     Args:
    #         repository (str): The repository to set for the tag
    #         tag (str): The tag name
    #         force (bool): Force
    #
    #     Raises:
    #         :py:class:`docker.errors.APIError`
    #             If the server returns an error.
    #
    #     Returns:
    #         (bool): ``True`` if successful
    #     """
    #     raise NotImplementedError
    #
    # def build_and_push_image(self, case, repository, tag=None, **kwargs):
    #     image = self._build_image(case)
    #     image = self._tag_image(image, repository, tag, **kwargs)
    #     self._push_image(repository, tag, **kwargs)



import json
import os
import shutil
import uuid
from pathlib import Path
from typing import List

from cookiecutter.main import cookiecutter

try:
    import git
except ImportError:
    # Adding this hint to save users the confusion of trying $pip install git
    raise ImportError("No module named git, please install the gitpython package")

import cadetrdm
from cadetrdm.repositories import ProjectRepo, OutputRepo
from cadetrdm.io_utils import write_lines_to_file, wait_for_user, init_lfs, test_for_lfs
import cadetrdm.templates.dockerfile_template as dockerfile_template

def initialize_repo(path_to_repo: str | Path, output_folder_name: (str | bool) = "output", gitignore: list = None,
                    gitattributes: list = None, output_repo_kwargs: dict = None, cookiecutter_template: str = None):
    """
    Initialize a git repository at the given path with an optional included output results repository.

    :param path_to_repo:
        Path to main repository.
    :param output_folder_name:
        Name for the output repository.
    :param gitignore:
        List of files to be added to the gitignore file.
    :param gitattributes:
        List of lines to be added to the gitattributes file
    :param output_repo_kwargs:
        kwargs to be given to the creation of the output repo initalization function.
        Include gitignore, gitattributes, and lfs_filetypes kwargs.
    """
    test_for_lfs()

    if cookiecutter_template is not None:
        init_cookiecutter(cookiecutter_template, path_to_repo)

    if gitignore is None:
        gitignore = get_default_gitignore() + ["*.ipynb", "*.h5"]

    gitignore.append(f"/{output_folder_name}/")
    gitignore.append(f"/{output_folder_name}_cached/")

    if gitattributes is not None:
        write_lines_to_file(path=".gitattributes", lines=gitattributes, open_type="a")

    if output_repo_kwargs is None:
        output_repo_kwargs = {}

    starting_directory = os.getcwd()
    project_repo_uuid = str(uuid.uuid4())
    output_repo_uuid = str(uuid.uuid4())

    if path_to_repo != ".":
        os.makedirs(path_to_repo, exist_ok=True)
        os.chdir(path_to_repo)

    initialize_git()

    write_lines_to_file(path=".gitignore", lines=gitignore, open_type="a")

    create_readme()
    create_environment_yml()
    create_dockerfile()

    ProjectRepo._add_jupytext_file()

    rdm_data = {
        "is_project_repo": True, "is_output_repo": False,
        "project_uuid": project_repo_uuid, "output_uuid": output_repo_uuid,
        "cadet_rdm_version": cadetrdm.__version__,
        "output_remotes": {"output_folder_name": output_folder_name, "output_remotes": {}}
    }
    with open(".cadet-rdm-data.json", "w") as f:
        json.dump(rdm_data, f, indent=2)

    with open(".cadet-rdm-cache.json", "w") as f:
        json.dump({"__example/path/to/repo__": {
            "source_repo_location": "git@jugit.fz-juelich.de:IBG-1/ModSim/cadet"
                                    "/agile_cadet_rdm_presentation_output.git",
            "branch_name": "output_from_master_3910c84_2023-10-25_00-17-23",
            "commit_hash": "6e3c26527999036e9490d2d86251258fe81d46dc"
        }}, f, indent=2)

    initialize_output_repo(output_folder_name, project_repo_uuid=project_repo_uuid,
                           output_repo_uuid=output_repo_uuid, **output_repo_kwargs)

    repo = ProjectRepo(".")
    repo.update_output_remotes_json()

    files = [".gitignore",
             "README.md",
             ".cadet-rdm-cache.json",
             ".cadet-rdm-data.json",
             "environment.yml",
             "jupytext.yml",
             "Dockerfile",
             ]
    if gitattributes is not None:
        files.append(".gitattributes")

    for file in files:
        repo._git.add(file)

    repo.commit("initial CADET RDM commit", add_all=False)

    os.chdir(starting_directory)
    del repo


def init_cookiecutter(cookiecutter_template, path_to_repo):
    """
    Initialize from cookiecutter template. Because cookiecutter can only create the files in a sub-directory
    but cadet-rdm init can be called from within a folder with "path_to_repo" == ".", we copy the files from the
    generated_dir folder into the path_to_repo folder afterwards.

    :param cookiecutter_template:
    :param path_to_repo:
    """
    generated_dir = cookiecutter(cookiecutter_template, output_dir=path_to_repo)
    file_names = os.listdir(generated_dir)
    for file_name in file_names:
        shutil.move(os.path.join(generated_dir, file_name), path_to_repo)
    shutil.rmtree(generated_dir)


# def re_initialize_existing_repo(path_to_repo: str | Path, **output_repo_kwargs):
#     path_to_repo = "."
#     output_repo_kwargs = {}
#
#     starting_directory = os.getcwd()
#     os.chdir(path_to_repo)
#
#     repo = ProjectRepo(".")
#
#     if Path(repo._output_folder).exists():
#         raise RuntimeError(f"Output repo at {repo._output_folder} already exists.")
#
#     initialize_output_repo(repo._output_folder, project_repo_uuid=repo._project_uuid,
#                            output_repo_uuid=repo._output_uuid, **output_repo_kwargs)
#
#     os.chdir(starting_directory)


def initialize_git(folder="."):
    starting_directory = os.getcwd()
    if folder != ":":
        os.chdir(folder)

    try:
        repo = git.Repo(".")
        proceed = wait_for_user('The target directory already contains a git repo.\n'
                                'Please commit or stash all changes to the repo before continuing.\n'
                                'Proceed?')
        if not proceed:
            raise KeyboardInterrupt
    except git.exc.InvalidGitRepositoryError:
        os.system(f"git init -b main")

    if folder != ":":
        os.chdir(starting_directory)


def get_default_gitignore():
    return [".idea", "*diskcache*", "*tmp*", ".ipynb_checkpoints", "__pycache__"]


def get_default_lfs_filetypes():
    return ["*.jpg", "*.png", "*.xlsx", "*.h5", "*.ipynb", "*.pdf", "*.docx", "*.zip", "*.html", "*.csv"]


def initialize_output_repo(output_folder_name, gitignore: list = None,
                           gitattributes: list = None, lfs_filetypes: list = None,
                           project_repo_uuid: str = None, output_repo_uuid: str = None):
    """
    Initialize a git repository at the given path with an optional included output results repository.

    :param output_folder_name:
        Name for the output repository.
    :param gitignore:
        List of files to be added to the gitignore file.
    :param gitattributes:
        List of lines to be added to the gitattributes file
    :param lfs_filetypes:
        List of filetypes to be handled by git lfs.
    """
    starting_directory = os.getcwd()
    os.makedirs(output_folder_name, exist_ok=True)
    os.chdir(output_folder_name)

    if gitignore is None:
        gitignore = get_default_gitignore()

    if gitattributes is None:
        gitattributes = ["log.tsv merge=union"]

    if lfs_filetypes is None:
        lfs_filetypes = get_default_lfs_filetypes()

    initialize_git()

    write_lines_to_file(path=".gitattributes", lines=gitattributes, open_type="a")
    write_lines_to_file(path=".gitignore", lines=gitignore, open_type="a")

    rdm_data = {
        "is_project_repo": False, "is_output_repo": True,
        "project_uuid": project_repo_uuid, "output_uuid": output_repo_uuid,
        "cadet_rdm_version": cadetrdm.__version__
    }
    with open(".cadet-rdm-data.json", "w") as f:
        json.dump(rdm_data, f, indent=2)

    init_lfs(lfs_filetypes)

    create_output_readme()

    repo = OutputRepo(".")
    repo.commit("initial commit")

    os.chdir(starting_directory)


def create_environment_yml():
    file_lines = ["name: rdm_example", "channels:", "  - conda-forge", "dependencies:", "  - python=3.10",
                  "  - cadet", "  - pip", "  - pip:", "      - cadet-process", "      - cadet-rdm"]
    if not os.path.exists("environment.yml"):
        write_lines_to_file("environment.yml", file_lines, open_type="w")


def create_readme():
    readme_lines = ["## Output Repository", 
                    "",
                    "The output data for this case study can be found here:",
                    "[Link to Output Repository]() (not actually set yet because no remote has been configured at this moment)"]
    write_lines_to_file("README.md", readme_lines, open_type="a")


def create_output_readme():
    readme_lines = ["# Output repository for Example Simulation with CADET",
                    "This repository stores the simulation results for RDM-Example. `CADET-RDM` automatically tracks all simulations that are started by running `main.py` from the corresponding project repository.",
                    "",
                    "Each simulation run creates a dedicated branch in this output repository. The results are saved within the `src` folder of the respective branch. Additionally, a `log.tsv` file in the main branch records metadata for all runs, uniquely linking each output branch to its originating run in the project repository.",
                    "",
                    "## Project Repository",
                    "",
                    "The project repository for this case study is available here: ",
                    "[Link to Project Repository]() (not actually set yet because no remote has been configured at this moment)"]
    write_lines_to_file("README.md", readme_lines, open_type="a")


def create_dockerfile():
    write_lines_to_file("Dockerfile", dockerfile_template, open_type="w")


import numpy as np
import re

from cadetrdm import Options
from cadetrdm.options import remove_invalid_keys
from cadetrdm import process_example
from cadetrdm import ProjectRepo
from pathlib import Path

def test_options_hash():
    opt = Options()
    opt["array"] = np.linspace(2, 200)
    opt["nested_dict"] = {"ba": "foo", "bb": "bar"}
    initial_hash = opt.get_hash()
    s = opt.dumps()
    opt_recovered = Options.loads(s)
    post_serialization_hash = opt_recovered.get_hash()
    assert initial_hash == post_serialization_hash
    assert opt == opt_recovered


def test_options_file_io():
    opt = Options()
    opt["array"] = np.linspace(0, 2, 200)
    opt["nested_dict"] = {"ba": "foo", "bb": "bar"}
    initial_hash = opt.get_hash()
    opt.dump_json_file("options.json")
    opt_recovered = Options.load_json_file("options.json")
    post_serialization_hash = opt_recovered.get_hash()
    assert initial_hash == post_serialization_hash
    assert opt == opt_recovered


def test_remove_keys_starting_with_underscore():
    input_dict = {
        "_private": 1,
        "valid": 2,
        "__magic__": 3
    }
    expected = {"valid": 2}
    assert remove_invalid_keys(input_dict) == expected


def test_remove_keys_containing_double_underscore():
    input_dict = {
        "normal": 1,
        "with__double": 2,
        "another": 3
    }
    expected = {"normal": 1, "another": 3}
    assert remove_invalid_keys(input_dict) == expected


def test_nested_dict_removal():
    input_dict = {
        "level1": {
            "_invalid": 1,
            "valid": {
                "__still_invalid__": 2,
                "ok": 3
            }
        },
        "__should_be_removed__": "nope"
    }
    expected = {
        "level1": {
            "valid": {
                "ok": 3
            }
        }
    }
    assert remove_invalid_keys(input_dict) == expected


def test_empty_dict():
    assert remove_invalid_keys({}) == {}


def test_all_invalid_keys():
    input_dict = {
        "_one": 1,
        "__two__": 2,
        "with__double": 3
    }
    assert remove_invalid_keys(input_dict) == {}


def test_no_invalid_keys():
    input_dict = {
        "a": 1,
        "b": {
            "c": 2
        }
    }
    assert remove_invalid_keys(input_dict) == input_dict

def test_explicit_invalid_keys():
    input_dict = {
        "a": 1,
        "b": {
            "c": 2
        }
    }
    expected = {
        "b": {
            "c": 2
        }
    }
    assert remove_invalid_keys(input_dict, excluded_keys=["a"]) == expected

def test_branch_name():
    options = Options()
    options.commit_message = "Commit Message Test"
    options.debug = True
    options.push = False
    options.source_directory = "src"

    repo = ProjectRepo(Path("./test_repo_cli"), options=options)

    hash = str(repo.head.commit)[:7]
    active_branch = str(repo.active_branch)
    new_branch = repo.get_new_output_branch_name()

    escaped_branch = re.escape(active_branch)

    pattern = rf"^\d{{4}}-\d{{2}}-\d{{2}}_\d{{2}}-\d{{2}}-\d{{2}}_{escaped_branch}_{hash}$"

    assert re.match(pattern, new_branch), f"Branch name '{new_branch}' does not match expected format"

def test_branch_name_with_prefix():

    options = Options()
    options.commit_message = "Commit Message Test"
    options.debug = True
    options.push = False
    options.source_directory = "src"
    options.branch_prefix = "Test_Prefix"

    repo = ProjectRepo(Path("./test_repo_cli"), options=options)

    hash = str(repo.head.commit)[:7]
    active_branch = str(repo.active_branch)
    new_branch = repo.get_new_output_branch_name()

    escaped_branch = re.escape(active_branch)

    pattern = rf"^Test_Prefix_\d{{4}}-\d{{2}}-\d{{2}}_\d{{2}}-\d{{2}}-\d{{2}}_{escaped_branch}_{hash}$"

    assert re.match(pattern, new_branch), f"Branch name '{new_branch}' does not match expected format"


import os
import shutil
from _stat import S_IWRITE
from pathlib import Path


def add_linebreaks(input_list, initial_linebreak=True):
    """
    Add linebreaks between each entry in the input_list

    :param input_list:
    List of strings to add linebreaks to.

    :param initial_linebreak:
    Bool, if true: add a newline before the first line.
    """
    lines = [line + "\n" for line in input_list]
    if initial_linebreak:
        lines = ["\n"] + lines
    return lines


def write_lines_to_file(path, lines, open_type="a"):
    """
    Convenience function. Write lines to a file at path with added newlines between each line.
    :param path:
        Path to file.
    :param lines:
        List of lines to be written to file.
    :param open_type:
        The way the file should be opened. I.e. "a" for append and "w" for fresh write.
    """

    add_initial_linebreak = False

    if os.path.exists(path) and open_type == "a":
        with open(path, "r", encoding="utf-8") as f:
            existing_lines = f.readlines()
        if len(existing_lines) > 0 and not existing_lines[-1].endswith("\n"):
            add_initial_linebreak = True

    with open(path, open_type, encoding="utf-8") as f:
        f.writelines(add_linebreaks(lines, initial_linebreak=add_initial_linebreak))


def is_tool(name):
    """Check whether `name` is on PATH and marked as executable."""

    from shutil import which
    return which(name) is not None


def recursive_chmod(path, setting):
    for dirpath, dirnames, filenames in os.walk(path):
        os.chmod(dirpath, setting)
        for filename in filenames:
            os.chmod(os.path.join(dirpath, filename), setting)


def delete_path(filename):
    def remove_readonly(func, path, exc_info):
        # Clear the readonly bit and reattempt the removal
        # ERROR_ACCESS_DENIED = 5
        if func not in (os.unlink, os.rmdir) or exc_info[1].winerror != 5:
            raise exc_info[1]
        os.chmod(path, S_IWRITE)
        func(path)

    absolute_path = os.path.abspath(filename)
    if os.path.isdir(absolute_path):
        shutil.rmtree(absolute_path, onerror=remove_readonly)
    else:
        os.remove(absolute_path)


def wait_for_user(message):
    proceed = input(message + " Y/n \n")
    if proceed.lower() == "y" or proceed == "":
        return True
    else:
        return False


def init_lfs(lfs_filetypes: list, path: str | Path = None):
    """
    Initialize lfs in the git repository at the path.
    If path is None, the current working directory is used.
    :param lfs_filetypes:
        List of file types to be handled by lfs.
        Format should be e.g. ["*.jpg", "*.png"] for jpg and png files.
    """
    if path is not None:
        previous_path = os.getcwd()
        os.chdir(path)
    else:
        previous_path = "."

    os.system(f"git lfs install")
    lfs_filetypes_string = " ".join(lfs_filetypes)
    os.system(f"git lfs track {lfs_filetypes_string}")

    if path is not None:
        os.chdir(previous_path)


def test_for_lfs():
    if not is_tool("git-lfs"):
        raise RuntimeError("Git LFS is not installed. Please install it via e.g. apt-get install git-lfs or the "
                           "instructions found below \n"
                           "https://docs.github.com/en/repositories/working-with-files"
                           "/managing-large-files/installing-git-large-file-storage")



from urllib.parse import urlparse
from pathlib import Path


def ssh_url_to_http_url(url):
    if "https" in url:
        return url
    if Path(url).exists():
        return url

    url = url.replace(":", "/").replace("git@", "https://").replace(".git", "")
    return url


def is_valid_url(string):
    try:
        result = urlparse(string)
        return all([result.scheme, result.netloc])
    except:
        return False



import os
import subprocess
import tempfile
from pathlib import Path

try:
    import docker
    from docker.models.images import Image
except ImportError:
    print("Warning: no python-docker-interface installation found.")
import yaml

from cadetrdm.container import ContainerAdapter
from cadetrdm.batch_running import Case
from cadetrdm import Environment, ProjectRepo, Options


class DockerAdapter(ContainerAdapter):

    def __init__(self):
        self.client = docker.from_env()
        self.image = None

    def run_yml(self, yml_path):
        with open(yml_path, "r") as stream:
            instructions = yaml.safe_load(stream)

        instructions = {key.lower(): value for key, value in instructions.items()}

        study = ProjectRepo(**instructions["projectrepo"], suppress_lfs_warning=True)
        options = Options(**instructions["options"])
        environment = Environment(**instructions["environment"])
        case = Case(study, options, environment)

        return self.run_case(case, command=instructions["command"])

    def run_case(self, case: Case, command: str = None):

        if case.environment is not None:
            self._update_Dockerfile_with_env_reqs(case)

        if self.image is None:
            image = self._build_image(case)
        else:
            image = self.image

        container_tmp_filename = "/tmp/options.json"
        options_tmp_filename = self._dump_options(case)

        full_command = self._prepare_case_command(
            case=case,
            command=command,
            container_options_filename=container_tmp_filename
        )

        log, return_code = self._run_command(
            full_command=full_command,
            image=image,
            mounts={options_tmp_filename: container_tmp_filename},
        )

        return log, return_code

    def _run_command(self, full_command, image, mounts=None):
        """

        :param full_command:
        :param image:
        :param mounts: Dictionary mapping host paths to container paths
        :return:
        """

        ssh_location = Path.home() / ".ssh"
        if not ssh_location.exists():
            raise FileNotFoundError("No ssh folder found. Please report this on GitHub/CADET/CADET-RDM")

        volumes = {
            f"{Path.home()}/.ssh": {'bind': "/root/.ssh_host_os", 'mode': "ro"},
        }
        if mounts is None:
            mounts = {}
        for host_path, container_path in mounts.items():
            volumes[host_path.absolute().as_posix()] = {'bind': container_path, 'mode': 'ro'}

        container = self.client.containers.run_yml(
            image=image,
            command=full_command,
            volumes=volumes,
            detach=True,
            remove=False
        )

        full_log = []
        # Step 2: Attach to the container's logs
        for log in container.logs(stream=True):
            full_log.append(log.decode("utf-8"))
            print(log.decode("utf-8"), end="")

        # Wait for the container to finish execution
        result = container.wait()
        exit_code = result["StatusCode"]

        container.remove()

        return full_log, exit_code

    def _dump_options(self, case):
        if not Path("tmp").exists():
            os.makedirs("tmp")
        tmp_filename = Path("tmp/" + next(tempfile._get_candidate_names()) + ".json")
        case.options.dump_json_file(tmp_filename)
        return tmp_filename

    def _build_image(self, case) -> Image:
        cwd = os.getcwd()
        with open(case.project_repo.path / "Dockerfile", "rb") as dockerfile:
            os.chdir(case.project_repo.path.as_posix())

            image, logs = self.client.images.build(
                path=case.project_repo.path.as_posix(),
                # fileobj=dockerfile,  # A file object to use as the Dockerfile.
                tag=case.project_repo.name + ":" + case.name[:10],  # A tag to add to the final image
                quiet=False,  # Whether to return the status
                pull=True,  # Downloads any updates to the FROM image in Dockerfiles

            )
        if case.options.debug:
            for log in logs:
                print(log)
        os.chdir(cwd)
        return image

    def pull_image(self, repository, tag=None, all_tags=False, **kwargs):
        self.image = self.client.images.pull(
            repository=repository,
            tag=tag,
            all_tags=all_tags,
            **kwargs
        )

    def _push_image(self, repository, tag=None, **kwargs):
        self.client.images.push(
            repository=repository,
            tag=tag,
            **kwargs
        )

    def _tag_image(self, image: Image, repository, tag=None, **kwargs) -> Image:
        """
        Tag this image into a repository. Similar to the ``docker tag``
        command.

        Args:
            repository (str): The repository to set for the tag
            tag (str): The tag name
            force (bool): Force

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.

        Returns:
            (bool): ``True`` if successful
        """
        image.tag(repository=repository, tag=tag, **kwargs)
        return image

    def build_and_push_image(self, case, repository, tag=None, **kwargs):
        image = self._build_image(case)
        image = self._tag_image(image, repository, tag, **kwargs)
        self._push_image(repository, tag, **kwargs)

    def _update_Dockerfile_with_env_reqs(self, case):
        case.project_repo._reset_hard_to_head(force_entry=True)

        dockerfile = Path(case.project_repo.path) / "Dockerfile"
        install_command = case.environment.prepare_install_instructions()
        if install_command is None:
            return

        with open(dockerfile, "a") as handle:
            handle.write(f"\n{install_command}\n")

    def __del__(self):
        self.client.close()



import gitlab
import github
import keyring
from abc import abstractmethod


class Remote:
    @staticmethod
    def load_token(url_options, username):
        token = None
        url_options_iter = iter(url_options)
        try:
            while token is None:
                token = keyring.get_password(next(url_options_iter), username)
        except StopIteration:
            raise RuntimeError(f"No token found in keyring for url {url_options} and username {username}")

        return token

    @abstractmethod
    def create_remote(self, url, namespace, name, username):
        return

    @abstractmethod
    def delete_remote(self, url, namespace, name, username):
        return


class GitLabRemote(Remote):

    @property
    def url_fallbacks(self):
        return ["gitlab"]

    def create_remote(self, url, namespace, name, username):
        """
        Create remotes on gitlab within the given url / namespace / name. Use the token
        stored in the keyring under the username and url combination.

        :param namespace:
        :param name:
        :param url:
        :param username:
        :return:
        Query response
        """
        namespace = namespace.lower()
        token = self.load_token([url] + self.url_fallbacks, username)
        gl = gitlab.Gitlab(url, private_token=token)
        # We need the groups list, because for some ineffable reason, gitlab doesn't always give all namespaces.
        # This assumes, that the group_id and the corresponding namespace_id are identical. So far this has been true.
        gl_groups = gl.groups.list(get_all=True)
        # We also need the namespace list for the personal namespace.
        gl_namespaces = gl.namespaces.list(get_all=True)
        matching_namespace = [gl_group for gl_group in gl_groups + gl_namespaces
                              if gl_group.full_path.lower() == namespace]

        if len(matching_namespace) == 0:
            raise ValueError(f"Could not find namespace {namespace} "
                             f"in {[gl_namespace.full_path for gl_namespace in gl_groups]}")
        if len(matching_namespace) >= 2:
            matching_namespace_id_set = set({group.id for group in matching_namespace})
            if len(matching_namespace_id_set) > 1:
                raise ValueError(f"Not unique namespace {namespace} "
                                 f"in {[gl_namespace.full_path.lower() for gl_namespace in gl_groups]}")

        namespace_id = matching_namespace[0].id

        response = gl.projects.create({"name": name, "namespace_id": namespace_id})
        return response

    def delete_remote(self, url, namespace, name, username):
        """
        Deletes remotes on gitlab within the given url / namespace / name. Use the token
        stored in the keyring under the username and url combination.

        :param namespace:
        :param name:
        :param url:
        :param username:
        :return:
        None
        """
        token = self.load_token([url] + self.url_fallbacks, username)
        gl = gitlab.Gitlab(url, private_token=token)

        potential_projects = gl.projects.list(get_all=True, search=[namespace, name])

        for project in potential_projects:
            if project.name != name:
                pass
            if project.namespace["name"] != namespace:
                pass

            gl.projects.delete(project.id)
        return


class GitHubRemote(Remote):

    @property
    def url_fallbacks(self):
        return ["https://github.com/", "https://github.com", "github", "github.com"]

    def create_remote(self, name, namespace=None, url="https://api.github.com", username=None):
        """
        Create remotes on GitHub within the given url / namespace / name. Use the token
        stored in the keyring under the username and url combination.

        :param namespace:
        :param name:
        :param url:
        :param username:
        :return:
        Query response
        """
        if username is None and namespace is not None:
            username = namespace

        token = self.load_token([url] + self.url_fallbacks, username)

        auth = github.Auth.Token(token)
        g = github.Github(base_url=url, auth=auth)
        user = g.get_user()

        if namespace is None or namespace == user.login:
            base = user
        else:
            try:
                organization = g.get_organization(namespace)
                base = organization
            except github.GithubException:
                raise RuntimeError(f"No organization or user named {namespace} found in {url}")

        response = base.create_repo(
            name,
            allow_rebase_merge=True,
            auto_init=False,
            has_issues=True,
            has_projects=False,
            has_wiki=False,
            private=False,
        )
        return response

    def delete_remote(self, name, namespace, url="https://api.github.com", username=None):
        if username is None:
            username = namespace

        token = self.load_token([url] + self.url_fallbacks, username)

        auth = github.Auth.Token(token)
        g = github.Github(base_url=url, auth=auth)
        repo = g.get_repo(f"{namespace}/{name}")
        repo.delete()



from .batch_running.study import Study
from .batch_running.case import Case

print("Deprecation Warning: Importing Study and Case from cadetrdm.batch_runner "
      "will be deprecated in the next mayor release. Please import from cadetrdm.")



import os

from git import Repo

from batch_runner import Options

n_commits = 60

os.chdir(
    r"C:\Users\ronal\PycharmProjects\benchmarking\tmp\puetmann2013\output"
)
repo = Repo(".")
repo.git.checkout("main")

filepath = "log.tsv"
with open(filepath) as handle:
    lines = handle.readlines()
lines = [line.replace("\n", "").split("\t") for line in lines]

# branches = repo.git.branch("-r").split("\n")
# branches = [branch.replace("  origin/", "") for branch in branches if "origin/main" not in branch]
# remote = repo.remotes[0]

all_options = {}
for line in lines[1:]:
    branch = line[1]
    repo.git.checkout(branch)
    if os.path.exists("options.json"):
        options = Options.load_json_file("options.json")
        try:
            options.pop("study_options")
        except:
            pass
        all_options[branch] = options
        print(line[-1], options.get_hash())
        line[-1] = options.get_hash()
    status = repo.git.status()
    if "Changes not staged for commit:" in status:
        os.system("git lfs fsck --pointers & git add . & git commit --amend --no-edit")

repo.git.checkout("main")
lines = ["\t".join(line) for line in lines]

with open(filepath, "w") as handle:
    handle.writelines("\n".join(lines))

# os.system("git add . & git commit --amend --no-edit")
os.system('git add . & git commit -m "remove study-options"')



import os
import random
from pathlib import Path

import git
import numpy as np
import pytest

from cadetrdm import initialize_repo, ProjectRepo, Options
from cadetrdm.initialize_repo import init_lfs
from cadetrdm.io_utils import delete_path
from cadetrdm.repositories import OutputRepo, BaseRepo
from cadetrdm.web_utils import ssh_url_to_http_url
from cadetrdm.wrapper import tracks_results


@pytest.fixture(scope="module")
def path_to_repo():
    # a "fixture" serves up shared, ready variables to test functions that should use the fixture as a kwarg
    return Path("test_repo")


def modify_code(path_to_repo):
    # Add changes to the project code
    random_number = random.randint(0, 265)
    filepath = path_to_repo / f"print_random_number.py"
    with open(filepath, "w") as file:
        file.write(f"print({random_number})\n")


def count_commit_number(repo):
    commit_log = repo._git.log("--oneline").split("\n")
    current_commit_number = len(commit_log)
    return current_commit_number


def example_generate_results_array(path_to_repo, output_folder):
    results_array = np.random.random((500, 3))
    np.savetxt(path_to_repo / output_folder / "result.csv", results_array, delimiter=",")
    return results_array


@tracks_results
def example_generate_results_with_options(repo, options):
    results_array = np.random.random((500, 3))
    np.savetxt(repo.output_path / "result.csv", results_array, delimiter=",")
    return results_array


def try_init_gitpython_repo(repo_path):
    Path(repo_path).exists()
    git.Repo(repo_path)
    return True


def try_initialize_git_repo(path_to_repo):
    if path_to_repo.exists():
        delete_path(path_to_repo)

    initialize_repo(path_to_repo, "results")

    assert try_init_gitpython_repo(path_to_repo)
    assert try_init_gitpython_repo(path_to_repo / "results")


def try_commit_code(path_to_repo):
    repo = ProjectRepo(path_to_repo)
    current_commit_number = count_commit_number(repo)

    modify_code(path_to_repo)
    repo.commit("add code to print random number", add_all=True)

    updated_commit_number = count_commit_number(repo)
    assert current_commit_number + 1 == updated_commit_number


def try_commit_code_without_code_changes(path_to_repo):
    repo = ProjectRepo(path_to_repo)
    current_commit_number = count_commit_number(repo)
    repo.commit("This commit will not be made", add_all=True)
    updated_commit_number = count_commit_number(repo)
    assert current_commit_number == updated_commit_number


def try_commit_results_data(path_to_repo):
    repo = ProjectRepo(path_to_repo)
    current_commit_number = count_commit_number(repo.output_repo)
    with repo.track_results(results_commit_message="Add array") as output_branch:
        example_generate_results_array(path_to_repo, output_folder=repo.output_path)
    updated_commit_number = count_commit_number(repo.output_repo)
    assert current_commit_number <= updated_commit_number
    assert str(repo.output_repo.active_branch) == output_branch
    return output_branch


def try_commit_results_with_options(path_to_repo):
    options = Options()
    options.commit_message = "test with options"
    options.debug = False
    options.example = "123"
    options.dont_push = True

    repo = ProjectRepo(path_to_repo)
    current_commit_number = count_commit_number(repo.output_repo)

    example_generate_results_with_options(options, repo_path=repo.path)

    updated_commit_number = count_commit_number(repo.output_repo)
    assert current_commit_number <= updated_commit_number

    return str(repo.output_repo.active_branch)


def try_output_function(path_to_repo):
    repo = ProjectRepo(path_to_repo)
    filepath = repo.output_path / "foo" / "bar"
    assert isinstance(filepath, Path)
    filepath = repo.output_data("foo/bar")
    assert isinstance(filepath, Path)


def try_print_log(path_to_repo):
    repo = ProjectRepo(path_to_repo)
    repo.print_output_log()


def try_commit_results_with_uncommitted_code_changes(path_to_repo):
    repo = ProjectRepo(path_to_repo)
    modify_code(path_to_repo)
    with pytest.raises(Exception):
        with repo.track_results(results_commit_message="Add array"):
            example_generate_results_array(path_to_repo, output_folder=repo.output_path)
    repo.commit("add code to print random number", add_all=True)


def try_load_previous_output(path_to_repo, branch_name):
    repo = ProjectRepo(path_to_repo)
    with repo.track_results(results_commit_message="Load array and extend"):
        cache_folder_path = repo.input_data(branch_name=branch_name)
        cached_array_path = cache_folder_path / "result.csv"
        previous_array = np.loadtxt(cached_array_path, delimiter=",")
        extended_array = np.concatenate([previous_array, previous_array])
        extended_array_file_path = path_to_repo / repo.output_path / "extended_result.csv"
        np.savetxt(extended_array_file_path,
                   extended_array,
                   delimiter=",")
        assert cached_array_path.exists()
        assert extended_array_file_path.exists()


def try_add_remote(path_to_repo):
    repo = ProjectRepo(path_to_repo)
    repo.add_remote("git@jugit.fz-juelich.de:IBG-1/ModSim/cadet/CADET-RDM.git")
    assert "origin" in repo._git_repo.remotes


def try_initialize_from_remote():
    if Path("test_repo_from_remote").exists():
        delete_path("test_repo_from_remote")
    ProjectRepo.clone(
        url="git@github.com:ronald-jaepel/rdm_testing_template.git",
        to_path="test_repo_from_remote"
    )
    assert try_init_gitpython_repo("test_repo_from_remote")

    repo = ProjectRepo("test_repo_from_remote")
    assert hasattr(repo, "output_path")


def test_init_over_existing_repo(monkeypatch):
    path_to_repo = Path("test_repo_init_over_repo")
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)
    os.chdir(path_to_repo)
    os.system(f"git init --initial-branch=master")
    with open("README.md", "w") as handle:
        handle.write("Readme-line 1\n")
    with open(".gitignore", "w") as handle:
        handle.write("foo.bar.*")
    repo = git.Repo(".")
    repo.git.add(".")
    repo.git.commit("-m", "Initial commit")
    os.chdir("..")

    # using monkeypath to simulate user input
    monkeypatch.setattr('builtins.input', lambda x: "Y")

    initialize_repo(path_to_repo)
    delete_path(path_to_repo)


def test_init_over_existing_code(monkeypatch):
    path_to_repo = Path("test_repo_init_over_code")
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)

    with open(path_to_repo / "foobar.py", "w") as handle:
        handle.write("print('Hello World')\n")
    with open(path_to_repo / ".gitignore", "w") as handle:
        handle.write("foo.bar.*")

    # using monkeypath to simulate user input
    monkeypatch.setattr('builtins.input', lambda x: "Y")

    initialize_repo(path_to_repo)
    repo = git.Repo(path_to_repo)
    assert "Untracked files:" in repo.git.status()

    delete_path(path_to_repo)


def test_cache_with_non_rdm_repo(monkeypatch):
    path_to_repo = Path("non_rdm_repo")
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)
    os.chdir(path_to_repo)
    os.system(f"git init")
    with open("README.md", "w") as handle:
        handle.write("Readme-line 1\n")
    with open(".gitignore", "w") as handle:
        handle.write("foo.bar.*")
    git_repo = git.Repo(".")
    git_repo.git.add(".")
    git_repo.git.commit("-m", "Initial commit")

    os.chdir("..")
    if Path("test_repo_non_rdm_imports").exists():
        delete_path("test_repo_non_rdm_imports")
    initialize_repo("test_repo_non_rdm_imports")
    os.chdir("test_repo_non_rdm_imports")

    base_repo = BaseRepo(".")

    if Path("external_cache/non_rdm_repo").exists():
        delete_path("external_cache/non_rdm_repo")
    if Path("foo/bar/non_rdm_repo").exists():
        delete_path("foo/bar/non_rdm_repo")
    # import two repos and confirm verify works.
    base_repo.import_remote_repo(source_repo_location=".." / path_to_repo, source_repo_branch="master")
    base_repo.import_remote_repo(source_repo_location=".." / path_to_repo, source_repo_branch="master",
                            target_repo_location="foo/bar/non_rdm_repo")
    base_repo.verify_unchanged_cache()
    os.chdir("..")


def test_add_lfs_filetype():
    path_to_repo = Path("test_repo_add_lfs")
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)
    initialize_repo(path_to_repo)
    file_type = "*.bak"
    repo = ProjectRepo(path_to_repo)
    repo.output_repo.add_filetype_to_lfs(file_type)
    with open(repo.output_path / ".gitattributes", "r") as handle:
        gittatributes = handle.readlines()
    assert any(file_type in line for line in gittatributes)


def test_rdm_check():
    path_to_repo = Path("test_repo_rdm_check")
    new_project_url = "git@github.com:foobar/rdm_example_alternate.git"
    new_output_url = "git@github.com:foobar/rdm_example_output_alternate.git"
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)
    initialize_repo(path_to_repo)
    repo = ProjectRepo(path_to_repo)
    repo.add_remote("git@github.com:foobar/rdm_testing_template.git")
    repo.output_repo.add_remote("git@github.com:foobar/rdm_testing_template_output.git")
    repo.remote_set_url("origin", new_project_url)
    repo.output_repo.remote_set_url("origin", new_output_url)

    repo.check()
    with open(repo.path / "README.md", "r") as handle:
        readme_lines = handle.readlines()
    assert f'[Link to Output Repository]({ssh_url_to_http_url(new_output_url)})\n' in readme_lines
    with open(repo.output_repo.path / "README.md", "r") as handle:
        readme_lines = handle.readlines()
    assert f'[Link to Project Repository]({ssh_url_to_http_url(new_project_url)})\n' in readme_lines


def test_copy_external_data():
    path_to_source = Path("test_repo_external_data_source")
    if path_to_source.exists():
        delete_path(path_to_source)
    os.makedirs(path_to_source)

    filepath = path_to_source / f"static_data_contents"
    with open(filepath, "w") as file:
        file.write("This is static data")

    path_to_repo = Path("test_repo_external_data")
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)
    initialize_repo(path_to_repo)
    repo = ProjectRepo(path_to_repo)
    modify_code(path_to_repo)
    branch_name = repo.import_static_data(
        "test_repo_external_data_source",
        "import non_rdm_repo"
    )
    assert repo.has_uncomitted_changes
    cache_path = repo.copy_data_to_cache(branch_name)
    assert (cache_path / "test_repo_external_data_source" / "static_data_contents").exists()


def test_error_stack():
    path_to_repo = Path("test_repo_errors")
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)
    initialize_repo(path_to_repo)
    repo = ProjectRepo(path_to_repo)
    try:
        with repo.track_results("test error"):
            raise ValueError("This is an error message with \n a line break")
    except ValueError:
        pass
    with open("test_repo_errors/output/error.stack", "r") as handle:
        lines = handle.readlines()

    error_line = '    raise ValueError("This is an error message with \\n a line break")\n'
    assert error_line in lines


def test_cookiecutter_with_url(monkeypatch):
    path_to_repo = Path("test_repo_cookie")
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)

    # using monkeypatch to simulate user input
    # Mock console.input to simulate user input
    from rich.console import Console
    monkeypatch.setattr(Console, "input", lambda *args, **kwargs: "")

    initialize_repo(
        path_to_repo,
        cookiecutter_template="https://github.com/cadet/RDM-Cookiecutter-Example-Template.git"
    )
    assert (path_to_repo / "main.py").exists()


def test_cadet_rdm(path_to_repo):
    # because these depend on one-another and there is no native support afaik for sequential tests
    # these tests are called sequentially here as try_ functions.
    try_initialize_git_repo(path_to_repo)
    try_initialize_from_remote()

    try_add_remote(path_to_repo)
    # try_add_submodule(path_to_repo)
    try_commit_code(path_to_repo)
    try_commit_code_without_code_changes(path_to_repo)
    try_commit_results_with_uncommitted_code_changes(path_to_repo)
    try_output_function(path_to_repo)

    try_commit_results_with_options(path_to_repo)
    results_branch_name = try_commit_results_data(path_to_repo)
    try_print_log(path_to_repo)

    try_commit_code(path_to_repo)

    try_load_previous_output(path_to_repo, results_branch_name)


def test_with_detached_head():
    path_to_repo = Path("test_repo_2")
    if path_to_repo.exists():
        delete_path(path_to_repo)
    os.makedirs(path_to_repo)
    initialize_repo(path_to_repo)
    os.chdir(path_to_repo)
    for i in range(2):
        with open("README.md", "a") as handle:
            handle.write(f"Readme-line {i}\n")
        os.system(f"git add .")
        os.system(f"git commit -m foobar{i}")

    os.system("git checkout HEAD~1")
    repo = ProjectRepo(".")
    with repo.track_results("foo"):
        pass

    os.chdir("..")


# def test_with_external_repos():
#     path_to_repo = Path("test_repo_external_data")
#     if path_to_repo.exists():
#         delete_path(path_to_repo)
#     os.makedirs(path_to_repo)
#     initialize_repo(path_to_repo)
#
#     os.chdir(path_to_repo)
#
#     # to be able to hand over a valid branch, I first need to determine that branch
#     imported_repo = OutputRepo("../test_repo/results")
#     branch_name = imported_repo.active_branch.name
#
#     repo = ProjectRepo(".")
#
#     # import two repos and confirm verify works.
#     repo.import_remote_repo(source_repo_location="../test_repo/results", source_repo_branch=branch_name)
#     repo.import_remote_repo(source_repo_location="../test_repo/results", source_repo_branch=branch_name,
#                             target_repo_location="foo/bar/repo")
#     # delete folder and reload
#     delete_path("foo/bar/repo")
#
#     with pytest.raises(Exception):
#         repo.verify_unchanged_cache()
#
#     repo.fill_data_from_cadet_rdm_json()
#     repo.verify_unchanged_cache()
#
#     # Test if re_load correctly re_loads by modifying and then reloading
#     with open("external_cache/results/README.md", "w") as file_handle:
#         file_handle.writelines(["a", "b", "c"])
#     repo.fill_data_from_cadet_rdm_json(re_load=True)
#     repo.verify_unchanged_cache()
#
#     # modify file and confirm error raised
#     with open("external_cache/results/README.md", "w") as file_handle:
#         file_handle.writelines(["a", "b", "c"])
#     with pytest.raises(Exception):
#         repo.verify_unchanged_cache()
#
#     os.chdir("..")



import pytest
from pathlib import Path
import random
import os

from click.testing import CliRunner

from cadetrdm.cli_integration import cli
from cadetrdm.io_utils import delete_path

runner = CliRunner()


def create_repo():
    if os.path.exists("test_repo_cli"):
        delete_path("test_repo_cli")
    os.makedirs("test_repo_cli", exist_ok=True)
    os.chdir("test_repo_cli")
    result = runner.invoke(cli, ["init", ])


def modify_code(path_to_repo):
    # Add changes to the project code
    random_number = random.randint(0, 265)
    filepath = Path(path_to_repo) / f"print_random_number.py"
    with open(filepath, "w") as file:
        file.write(
            'with open("output/data.txt", "w") as handle:\n'
            f"    handle.write('{random_number}')\n"
        )


def test_01_initialize_repo():
    try:
        if os.path.exists("test_repo_cli"):
            delete_path("test_repo_cli")
        os.makedirs("test_repo_cli", exist_ok=True)

        os.chdir("test_repo_cli")
        result = runner.invoke(cli, ["init", ])
        print(result.output)
        assert result.exit_code == 0
    finally:
        os.chdir("..")


def test_02_add_remote():
    try:
        create_repo()
        result = runner.invoke(cli, ["remote", "add", "https://jugit.fz-juelich.de/r.jaepel/API_test_project"])
        print(result.output)
        assert result.exit_code == 0
        os.chdir("output")
        result = runner.invoke(cli, ["remote", "add", "https://jugit.fz-juelich.de/r.jaepel/API_test_project_output"])
        print(result.output)
        os.chdir("..")
        assert result.exit_code == 0
    finally:
        os.chdir("..")


@pytest.mark.server_api
def test_02b_clone():
    if os.path.exists("test_repo_cli_cloned"):
        delete_path("test_repo_cli_cloned")
    result = runner.invoke(cli, ["clone", "test_repo_cli", "test_repo_cli_cloned"])
    print(result.output)
    assert result.exit_code == 0


def test_03_commit_results_with_uncommited_code_changes():
    try:
        create_repo()
        modify_code(".")

        result = runner.invoke(cli, ["run_yml", "python", "print_random_number.py",
                                     "create data"])
        print(result.output)
        assert result.exit_code != 0
    finally:
        os.chdir("..")


def test_04_commit_code():
    try:
        create_repo()
        modify_code(".")

        result = runner.invoke(cli, ["commit", "-m", "add code", "-a"])
        print(result.output)
        assert result.exit_code == 0
    finally:
        os.chdir("..")


# def test_05_commit_results():
#     result = runner.invoke(cli, ["commit", "-m", "add code", "-a"])
#     print(result.output)
#     assert result.exit_code == 0
#     result = runner.invoke(cli, ["run_yml", "python", "print_random_number.py",
#                                  "create data"])
#     print(result.output)
#     assert result.exit_code == 0
#

def test_05b_execute_command():
    try:
        create_repo()
        modify_code(".")
        result = runner.invoke(cli, ["commit", "-a", "-m", "add code"])
        print(result.output)
        assert result.exit_code == 0

        filepath = Path(".") / f"print_random_number.py"
        result = runner.invoke(cli, ["run", "command", f"python {filepath.absolute().as_posix()}",
                                     "create data"])
        print(result.output)
        assert result.exit_code == 0
    finally:
        os.chdir("..")


def test_06_print_log():
    try:
        create_repo()
        result = runner.invoke(cli, ["log"])
        print(result.output)
        assert result.exit_code == 0
    finally:
        os.chdir("..")


def test_07_lfs_add():
    try:
        create_repo()
        result = runner.invoke(cli, ["lfs", "add", "pptx"])
        print(result.output)
        assert result.exit_code == 0
    finally:
        os.chdir("..")


def test_08_data_import():
    try:
        create_repo()
        result = runner.invoke(cli,
                               [
                                   "data", "clone",
                                   "git@github.com:cadet/CADET-Core.git",
                                   "master",
                                   "imported/repo/data"
                               ])
        print(result.stdout)
        assert result.exit_code == 0
    finally:
        os.chdir("..")


@pytest.mark.container
def test_run_dockered():
    try:
        create_repo()
        result = runner.invoke(
            cli,
            ["run_yml", "dockered", (Path(__file__).parent.resolve() / "case.yml").as_posix()]
        )
        print(result.output)
        assert result.exit_code == 0
    finally:
        os.chdir("..")

# def test_09_data_verify():
#     with open()
#     result = runner.invoke(cli, ["data", "verify"])



from time import sleep

import git
import pytest

from cadetrdm import initialize_repo, ProjectRepo
from cadetrdm.io_utils import delete_path
from cadetrdm.remote_integration import GitHubRemote, GitLabRemote
from cadetrdm.repositories import BaseRepo


@pytest.mark.server_api
def test_gitlab_create():
    url = "https://jugit.fz-juelich.de/"
    namespace = "r.jaepel"
    name = "API_test_project"
    remote = GitLabRemote()

    # ensure remote does not exist
    remote.delete_remote(url=url, namespace=namespace, name=name, username="r.jaepel")
    try:
        delete_path("test_repo_remote")
    except FileNotFoundError:
        pass

    sleep(3)

    response = remote.create_remote(url=url, namespace=namespace, name=name, username="r.jaepel")

    BaseRepo.clone(response.ssh_url_to_repo, "test_repo_remote")
    delete_path("test_repo_remote")

    remote.delete_remote(url=url, namespace=namespace, name=name, username="r.jaepel")

    sleep(3)

    with pytest.raises(git.exc.GitCommandError):
        BaseRepo.clone(response.ssh_url_to_repo, "test_repo_remote")


@pytest.mark.server_api
def test_github_create():
    namespace = "ronald-jaepel"
    name = "API_test_project"
    remote = GitHubRemote()

    # ensure remote does not exist
    try:
        remote.delete_remote(namespace=namespace, name=name, username="r.jaepel")
    except Exception:
        pass

    try:
        delete_path("test_repo_remote")
    except FileNotFoundError:
        pass

    sleep(3)

    response = remote.create_remote(namespace=namespace, name=name, username="r.jaepel")

    sleep(3)

    BaseRepo.clone(response.html_url, "test_repo_remote")
    delete_path("test_repo_remote")

    remote.delete_remote(namespace=namespace, name=name, username="r.jaepel")

    with pytest.raises(git.exc.GitCommandError):
        BaseRepo.clone(response.ssh_url, "test_repo_remote")


@pytest.mark.server_api
def test_repo_gitlab_integration():
    url = "https://jugit.fz-juelich.de/"
    namespace = "r.jaepel"
    name = "API_test_project"
    repo_name = "test_repo_remote"
    remote = GitLabRemote()

    # Clean up
    remote.delete_remote(url=url, namespace=namespace, name=name, username="r.jaepel")
    remote.delete_remote(url=url, namespace=namespace, name=name + "_output", username="r.jaepel")

    try:
        delete_path("test_repo_remote")
    except FileNotFoundError:
        pass

    initialize_repo(repo_name)

    repo = ProjectRepo(repo_name)
    repo.create_remotes(url=url, namespace=namespace, name=name, username="r.jaepel")

    assert repo.has_changes_upstream is False



import warnings

from cadetrdm import ProjectRepo


class Study(ProjectRepo):
    def __init__(self, path, url=None, branch=None, name=None, suppress_lfs_warning=False, *args, **kwargs):
        super().__init__(path=path, url=url, suppress_lfs_warning=suppress_lfs_warning, branch=branch, *args, **kwargs)
        warnings.warn(
            "cadetrdm.Study() will be deprecated soon. Please use ProjectRepo()",
            FutureWarning
        )



import io
import subprocess

import pytest

from cadetrdm import Environment
from cadetrdm.logging import OutputLog


def test_environment_from_yml():
    yml_string = (
        'name: cadetrdmpy3.11\n'
        'channels:\n  - conda-forge\n'
        'dependencies:\n'
        '  - bzip2=1.0.8=hcfcfb64_5\n'
        '  - ca-certificates=2024.12.14=h56e8100_0\n  - cadet=4.4.0=hdf1ca3b_1\n'
        '  - hdf5=1.14.3=nompi_h73e8ff5_100\n  - intel-openmp=2024.0.0=h57928b3_49841\n'
        '  - krb5=1.21.2=heb0366b_0\n  - libaec=1.1.2=h63175ca_1\n  - libblas=3.9.0=21_win64_mkl\n'
        '  - libcblas=3.9.0=21_win64_mkl\n  - libcurl=8.5.0=hd5e4a3a_0\n  - libexpat=2.6.1=h63175ca_0\n'
        '  - libffi=3.4.2=h8ffe710_5\n  - libhwloc=2.9.3=default_haede6df_1009\n  - libiconv=1.17=hcfcfb64_2\n'
        '  - liblapack=3.9.0=21_win64_mkl\n  - liblapacke=3.9.0=21_win64_mkl\n'
        '  - libsqlite=3.45.2=hcfcfb64_0\n  - libssh2=1.11.0=h7dfc565_0\n'
        '  - libxml2=2.12.5=hc3477c8_0\n  - libzlib=1.2.13=hcfcfb64_5\n  - mkl=2024.0.0=h66d3029_49657\n'
        '  - openssl=3.4.0=ha4e3fda_1\n  - pip=24.0=pyhd8ed1ab_0\n  - pthreads-win32=2.9.1=hfa6e2cd_3\n'
        '  - python=3.11.8=h2628c8c_0_cpython\n  - python_abi=3.11=5_cp311\n'
        '  - suitesparse=5.4.0=h5d0cbe0_1\n  - tbb=2021.11.0=h91493d7_1\n  - tbb-devel=2021.11.0=h3ec46f0_1\n'
        '  - tk=8.6.13=h5226925_1\n  - ucrt=10.0.22621.0=h57928b3_0\n  - vc=14.3=hcf57466_18\n'
        '  - vc14_runtime=14.38.33130=h82b7239_18\n  - vs2015_runtime=14.38.33130=hcb4865c_18\n'
        '  - wheel=0.42.0=pyhd8ed1ab_0\n  - xz=5.2.6=h8d14728_0\n  - yaml=0.2.5=h8ffe710_2\n'
        '  - pip:\n'
        '      - about-time==4.2.1\n      - addict==2.3.0\n      - alive-progress==3.1.5\n'
        '      - annotated-types==0.6.0\n      - anyio==4.3.0\n      - appdirs==1.4.4\n'
        '      - arviz==0.17.0\n      - asttokens==2.4.1\n      - attrs==23.2.0\n      - autograd==1.6.2\n'
        '      - betterproto==2.0.0b6\n      - build==1.1.1\n      - cadet-process==0.9.1\n'
        '      - cadet-python==0.14.1\n      - cadet-rdm==0.0.41\n      - certifi==2024.2.2\n'
        '      - cffi==1.16.0\n      - cfgv==3.4.0\n      - charset-normalizer==3.3.2\n'
        '      - click==8.1.7\n      - cma==3.2.2\n      - cobra==0.29.0\n      - colorama==0.4.6\n'
        '      - contourpy==1.2.0\n      - corner==2.2.2\n      - cryptography==42.0.5\n'
        '      - cycler==0.12.1\n      - decorator==5.1.1\n      - depinfo==2.2.0\n'
        '      - deprecated==1.2.14\n      - dill==0.3.8\n      - diskcache==5.6.3\n'
        '      - distlib==0.3.8\n      - docker==7.1.0\n      - executing==2.0.1\n'
        '      - filelock==3.13.1\n      - flake8==7.0.0\n      - fonttools==4.49.0\n'
        '      - future==1.0.0\n      - git-lfs==1.6\n      - gitdb==4.0.11\n      - gitpython==3.1.42\n'
        '      - grapheme==0.6.0\n      - grpclib==0.4.7\n      - h11==0.14.0\n      - h2==4.1.0\n'
        '      - h5netcdf==1.3.0\n      - h5py==3.10.0\n      - hagelkorn==1.2.3\n      - hopsy==1.4.1\n'
        '      - hpack==4.0.0\n      - httpcore==1.0.4\n      - httpx==0.27.0\n      - hyperframe==6.0.1\n'
        '      - identify==2.5.35\n      - idna==3.6\n      - importlib-metadata==7.0.2\n'
        '      - importlib-resources==6.3.0\n      - iniconfig==2.0.0\n      - ipython==8.25.0\n'
        '      - jaraco-classes==3.3.1\n      - jedi==0.19.1\n      - joblib==1.3.2\n'
        '      - keyring==24.3.1\n      - kiwisolver==1.4.5\n      - llvmlite==0.42.0\n'
        '      - markdown-it-py==3.0.0\n      - matplotlib==3.8.3\n      - matplotlib-inline==0.1.7\n'
        '      - mcbackend==0.5.2\n      - mccabe==0.7.0\n      - mdurl==0.1.2\n'
        '      - more-itertools==10.2.0\n      - mpmath==1.3.0\n      - multidict==6.0.5\n'
        '      - multiprocess==0.70.16\n      - nodeenv==1.8.0\n      - numba==0.59.0\n'
        '      - numpy==1.26.4\n      - optlang==1.8.1\n      - packaging==24.0\n      - pandas==2.2.1\n'
        '      - parso==0.8.4\n      - pathos==0.3.2\n      - pillow==10.2.0\n      - platformdirs==4.2.0\n'
        '      - pluggy==1.4.0\n      - polyround==0.2.11\n      - pox==0.3.4\n      - ppft==1.7.6.8\n'
        '      - pre-commit==3.6.2\n      - prompt-toolkit==3.0.47\n      - psutil==6.0.0\n'
        '      - pure-eval==0.2.2\n      - pycodestyle==2.11.1\n      - pycparser==2.21\n'
        '      - pydantic==2.6.4\n      - pydantic-core==2.16.3\n      - pyflakes==3.2.0\n'
        '      - pygithub==2.2.0\n      - pygments==2.17.2\n      - pyjwt==2.8.0\n      - pymoo==0.6.1.1\n'
        '      - pynacl==1.5.0\n      - pyparsing==3.1.2\n      - pypiwin32==223\n'
        '      - pyproject-hooks==1.0.0\n      - pyqt5==5.15.10\n      - pyqt5-qt5==5.15.2\n'
        '      - pyqt5-sip==12.13.0\n      - pytest==8.1.1\n      - python-benedict==0.33.2\n'
        '      - python-dateutil==2.9.0.post0\n      - python-fsutil==0.13.1\n      - python-gitlab==4.4.0\n'
        '      - python-libsbml==5.20.2\n      - python-slugify==8.0.4\n      - pytz==2024.1\n'
        '      - pywin32==306\n      - pywin32-ctypes==0.2.2\n      - pyyaml==6.0.1\n'
        '      - requests==2.31.0\n      - requests-toolbelt==1.0.0\n      - rich==13.7.1\n'
        '      - ruamel-yaml==0.18.6\n      - ruamel-yaml-clib==0.2.8\n      - scikit-learn==1.4.1.post1\n'
        '      - scipy==1.12.0\n      - setuptools==69.2.0\n      - six==1.16.0\n      - smmap==5.0.1\n'
        '      - sniffio==1.3.1\n      - stack-data==0.6.3\n      - swiglpk==5.0.10\n      - sympy==1.12\n'
        '      - tabulate==0.9.0\n      - text-unidecode==1.3\n      - threadpoolctl==3.3.0\n'
        '      - tqdm==4.66.2\n      - traitlets==5.14.3\n      - typing-extensions==4.10.0\n'
        '      - tzdata==2024.1\n      - urllib3==2.2.1\n      - virtualenv==20.25.1\n'
        '      - wcwidth==0.2.13\n      - wrapt==1.16.0\n      - xarray==2024.2.0\n'
        '      - xarray-einstats==0.7.0\n      - zipp==3.18.0\n'
        'prefix: C:\\Users\\ronal\\mambaforge\\envs\\cadetrdmpy3.11\n'
    )

    environment = Environment.from_yml_string(yml_string)
    assert environment.package_version("cadet") == "4.4.0"
    assert environment.fulfils("cadet", "~4.4.0")
    assert not environment.fulfils("cadet", ">4.4.0")
    assert environment.fulfils_environment(environment)

    requirements = {
        "text-unidecode": "1.3",
        "xarray": "<=2024.2.0",
        "zipp": "==3.18"
    }
    fulfilled_requirements = Environment(conda_packages={"cadet": ">=4.4.0"}, pip_packages=requirements)
    fulfilled_requirements.pip_packages["scikit-learn"] = "~1.4.1"

    assert environment.fulfils_environment(fulfilled_requirements)

    instructions = ("conda install -y cadet>=4.4.0 && "
                    "pip install 'text-unidecode==1.3' 'xarray<=2024.2.0' 'zipp==3.18' 'scikit-learn=1.4.1'")
    assert instructions == fulfilled_requirements.prepare_install_instructions()

    not_fulfilled_requirements = Environment(conda_packages={"cadet": ">=4.4.0"}, pip_packages={"xarray": ">=2026.2.0"})

    assert not environment.fulfils_environment(not_fulfilled_requirements)

    # test that the no-requirements case is always fulfilled.
    assert environment.fulfils_environment(None)

    # test eval() reproducibility
    assert environment.fulfils_environment(eval(repr(environment)))
    assert eval(repr(environment)).fulfils_environment(environment)


def test_environment():
    environment = Environment(conda_packages={"cadet": "4.4.0", "tbb": "2024.0.0"}, pip_packages={"xarray": "2024.2.0"})
    assert environment.package_version("cadet") == "4.4.0"
    assert environment.fulfils("cadet", "~4.4.0")
    assert not environment.fulfils("cadet", ">4.4.0")

    install_instructions = "conda install -y cadet=4.4.0 tbb=2024.0.0 && pip install 'xarray==2024.2.0'"
    assert environment.prepare_install_instructions() == install_instructions

    yml_dict = {'name': None,
                'channels': None,
                'dependencies': ['cadet=4.4.0',
                                 'tbb=2024.0.0',
                                 {'pip': ['xarray==2024.2.0']}]}
    assert environment._to_yml_dict() == yml_dict
    yml_string = ('channels: null\ndependencies:\n- cadet=4.4.0\n- tbb=2024.0.0\n'
                  '- pip:\n  - xarray==2024.2.0\nname: null\n')
    handle = io.StringIO()
    environment.to_yml(handle)
    assert yml_string == handle.getvalue()

    environment = Environment(conda_packages=None, pip_packages={"xarray": "2024.2.0"})
    yml_dict = {'name': None,
                'channels': None,
                'dependencies': [{'pip': ['xarray==2024.2.0']}]}
    assert environment._to_yml_dict() == yml_dict

    environment = Environment(conda_packages={"cadet": "4.4.0", "tbb": "2024.0.0"}, pip_packages=None)
    yml_dict = {'name': None,
                'channels': None,
                'dependencies': ['cadet=4.4.0', 'tbb=2024.0.0']}
    assert environment._to_yml_dict() == yml_dict

    complex_environment = Environment(
        conda_packages={"cadet": ">4.4.0", "tbb": ">=2024.0.0", "mkl": "~2024.0.0"},
        pip_packages={"xarray": ">2024.2.0", "pydantic": "<=2.6.4", }
    )

    install_instructions = ("conda install -y cadet>4.4.0 tbb>=2024.0.0 mkl=2024.0.0 && "
                            "pip install 'xarray>2024.2.0' 'pydantic<=2.6.4'")

    assert complex_environment.prepare_install_instructions() == install_instructions


@pytest.mark.slow
def test_update_environment():
    subprocess.run(f'conda env remove -n testing_env_cadet_rdm  -y', shell=True)
    subprocess.run(f'conda create -n testing_env_cadet_rdm python=3.11 -y', shell=True)
    target_env = Environment(
        conda_packages={"libiconv": "1.17", "openssl": ">=3.3"},
        # Currently not aware of any way of activating a conda env and then running commands in it. Therefore,
        # pip requirements are excluded from this test.
        # pip_packages={"cadet-rdm": "0.0.44"}
    )

    check = subprocess.run("conda env export -n testing_env_cadet_rdm ", shell=True, capture_output=True)
    current_env = Environment.from_yml_string(check.stdout.decode())
    assert not current_env.fulfils_environment(target_env)

    install_instructions = target_env.prepare_install_instructions()
    install_instructions = install_instructions.replace("install -y", "install --name testing_env_cadet_rdm -y")
    print(install_instructions)
    subprocess.run(install_instructions, shell=True)

    check = subprocess.run("conda env export -n testing_env_cadet_rdm ", shell=True, capture_output=True)
    current_env = Environment.from_yml_string(check.stdout.decode())
    assert current_env.fulfils_environment(target_env)
